from unittest import mock


from py_aws_core.testing import BaseTestFixture
from py_aws_core.spoofing.proxyrack import const
from py_aws_core.spoofing.proxyrack.backends import ProxyRackProxyBackend


class ProxyRackProxyBackendTests(BaseTestFixture):
    """
        ProxyRackProxyBackend Tests
    """

    @mock.patch.object(ProxyRackProxyBackend, 'get_proxy_password')
    @mock.patch.object(ProxyRackProxyBackend, 'get_proxy_username')
    def test_proxy_url(self, mocked_get_proxy_username, mocked_get_proxy_password):
        mocked_get_proxy_username.return_value = 'user123'
        mocked_get_proxy_password.return_value = 'pass456'
        proxy_url = ProxyRackProxyBackend.get_proxy_url(
            cities=['Dallas'],
            netloc='megaproxy.rotating.proxyrack.net:10000',
            proxy_ip='192.168.86.250',
            proxy_os=const.ProxyOs.WINDOWS,
            session_id='user123',
        )

        self.assertEqual(
            proxy_url,
            'http://user123;city=Dallas;osName=Windows;proxyIp=192.168.86.250;session=user123;refreshMinutes=60:pass456@megaproxy.rotating.proxyrack.net:10000'
        )



import json
import typing

from . import cookies, exceptions, logs, utils

logger = logs.logger


class LambdaEvent:
    class MultiValueHeaders:
        def __init__(self, data: dict):
            self.accept = data['Accept'][0]
            self.accept_encoding = data['Accept-Encoding'][0]
            self._authorization = data.get('Authorization')
            self._cookies = data.get('Cookie')
            self.user_agent = data['User-Agent'][0]

        @property
        def authorization(self):
            if self._authorization:
                return self._authorization[0]
            return None

        @property
        def cookies(self) -> typing.Dict:
            if not self._cookies:
                return dict()
            val = dict()
            for part in self._cookies[0].split(';'):
                k, v = part.strip().split('=')
                val[k] = v
            return val

    class RequestContext:
        def __init__(self, data):
            self.resource_id = data['resourceId']
            self.resource_path = data['resourcePath']
            self.http_method = data['httpMethod']
            self.request_time = data['requestTime']
            self.path = data['path']
            self.domain_name = data['domainName']

    def __init__(self, data):
        self._body = data['body']
        self.headers = data['headers'] or dict()
        self.http_method = data['httpMethod']
        self.multi_value_headers = self.MultiValueHeaders(data['multiValueHeaders'])
        self.multi_value_query_string_parameters = data['multiValueQueryStringParameters']
        self.path = data['path']
        self.query_string_parameters = data['queryStringParameters'] or dict()
        self.request_context = self.RequestContext(data['requestContext'])

    @property
    def body(self):
        if self._body:
            return json.loads(utils.remove_whitespace(self._body))
        return self._body

    @property
    def cookies(self) -> typing.Dict:
        return self.multi_value_headers.cookies

    @property
    def lower_headers(self) -> typing.Dict:
        return {k.lower(): v for k, v in self.headers.items()}

    def get_cookie(self, cookie_name: str):
        if cookie := self.cookies.get(cookie_name):
            logger.info(f'Cookie "{cookie_name}" found: {cookie}')
            return cookie
        raise exceptions.MissingCookieException(missing_cookie=cookie_name)

    def generate_cookie_header(self, name: str, value: str, expires_in_seconds: int) -> str:
        return cookies.CookieUtil.build_set_cookie_header(
            name=name,
            domain=self.request_context.domain_name,
            value=value,
            path='/',
            expires_in_seconds=expires_in_seconds
        )



from py_aws_core import logs
from py_aws_core.db_dynamo import get_db_client
from py_aws_core.db_session import SessionDBAPI

logger = logs.logger
db_client = get_db_client()


def rehydrate_session_from_database(client):
    session_id = client.session_id
    logger.info(f'Session ID: {session_id} -> Rehydrating session...')
    r_session = SessionDBAPI.GetSessionQuery.call(db_client=db_client, _id=session_id)
    if not r_session.sessions:
        logger.info(f'Session ID: {session_id} -> No prior session found.')
        return
    client.b64_decode_and_set_cookies(b64_cookies=r_session.session_b64_cookies)
    logger.info(f'Session ID: {session_id} -> Rehydrated {len(client.cookies.jar)} cookies')


def write_session_to_database(client):
    session_id = client.session_id
    logger.info(f'Session ID: {session_id} -> Writing cookies to database...')
    b64_cookies = client.b64_encoded_cookies
    c_maps = [SessionDBAPI.build_session_map(_id=client.session_id, b64_cookies=b64_cookies)]
    db_client.write_maps_to_db(item_maps=c_maps)
    logger.info(f'Session ID: {session_id} -> Wrote cookies to database')



import typing
import uuid

from py_aws_core import const, decorators, db_dynamo, entities, exceptions, logs
from py_aws_core.db_dynamo import DDBClient

logger = logs.logger


class SessionDBAPI(db_dynamo.ABCCommonAPI):
    @classmethod
    def build_session_map(
        cls,
        _id: uuid.UUID,
        b64_cookies: bytes,
        expire_in_seconds: int = const.DB_DEFAULT_EXPIRES_IN_SECONDS
    ):
        pk = sk = entities.Session.create_key(_id=_id)
        return cls.get_batch_entity_create_map(
            expire_in_seconds=expire_in_seconds,
            pk=pk,
            sk=sk,
            _type=entities.Session.type(),
            Base64Cookies=b64_cookies,
        )

    class GetSessionQuery:
        class Response(db_dynamo.QueryResponse):
            @property
            def sessions(self) -> typing.List:
                return self.get_by_type(entities.Session.type())

            @property
            def session_b64_cookies(self):
                return self.sessions[0]['Base64Cookies']['B']

        @classmethod
        @decorators.dynamodb_handler(client_err_map=exceptions.ERR_CODE_MAP, cancellation_err_maps=[])
        def call(
            cls,
            db_client: DDBClient,
            _id: uuid.UUID,
        ) -> Response:
            pk = entities.Session.create_key(_id=_id)
            response = db_client.query(
                key_condition_expression="#pk = :pk",
                expression_attribute_names={
                    "#pk": "PK",
                    "#cookies": "Base64Cookies",
                    "#typ": "Type"
                },
                expression_attribute_values={
                    ":pk": {"S": pk},
                },
                projection_expression='#cookies, #typ'
            )
            logger.debug(f'{cls.__qualname__}.call#: response: {response}')
            return cls.Response(response)

    class UpdateSessionCookies:
        @classmethod
        @decorators.dynamodb_handler(client_err_map=exceptions.ERR_CODE_MAP, cancellation_err_maps=[])
        def call(
            cls,
            db_client: DDBClient,
            _id: uuid.UUID,
            b64_cookies: bytes
        ):
            pk = sk = entities.Session.create_key(_id=_id)
            return db_client.update_item(
                key={
                    'PK': {'S': pk},
                    'SK': {'S': sk},
                },
                update_expression='SET Base64Cookies = :b64, ModifiedAt = :mda',
                expression_attribute_values={
                    ':b64': {'B': b64_cookies},
                    ':mda': {'S': SessionDBAPI.iso_8601_now_timestamp()}
                }
            )



import base64
import binascii
import pickle
import ssl
import uuid
from http.cookiejar import CookieJar

from httpx import Client, HTTPStatusError, TimeoutException, NetworkError, ProxyError

from py_aws_core import decorators, exceptions, logs, services, utils

logger = logs.logger
# Using same ssl context for all clients to save on loading SSL bundles
# See https://github.com/python/cpython/issues/95031#issuecomment-1749489998
# Also results in _tests running about 9 times faster
SSL_CONTEXT = ssl.create_default_context()


class RetryClient(Client):
    """
    Http/2 Client that retries for given exceptions and http status codes
    """
    RETRY_EXCEPTIONS = (
        HTTPStatusError,
        TimeoutException,
        NetworkError,
        ProxyError
    )

    RETRY_STATUS_CODES = (
        408,
        425,
        429,
        500,
        502,
        503,
        504,
    )

    def __init__(self, session_id: uuid.UUID = None, follow_redirects: bool = True, verify: bool = None, *args, **kwargs):
        super().__init__(
            follow_redirects=follow_redirects,
            default_encoding="utf-8",
            verify=verify or SSL_CONTEXT,
            *args,
            **kwargs
        )
        self._session_id = session_id or utils.get_uuid()

    @decorators.retry(retry_exceptions=RETRY_EXCEPTIONS)
    @decorators.http_status_check(reraise_status_codes=RETRY_STATUS_CODES)
    def send(self, *args, **kwargs):
        return super().send(*args, **kwargs)

    @property
    def session_id(self) -> uuid.UUID:
        return self._session_id

    @property
    def b64_encoded_cookies(self) -> bytes:
        return base64.encodebytes(pickle.dumps([c for c in self.cookies.jar]))

    def b64_decode_and_set_cookies(self, b64_cookies: bytes):
        if not b64_cookies:
            logger.info(f'Session ID: {self.session_id} -> No Cookies To Restore: {b64_cookies}')
            self.cookies.jar = CookieJar()
            return
        try:
            cookie_jar = CookieJar()
            decoded_bytes = base64.decodebytes(b64_cookies)
            for c in pickle.loads(decoded_bytes):
                logger.info(f'Session ID: {self.session_id} -> Setting CookieJar Cookie: "{c}"')
                cookie_jar.set_cookie(c)
            self.cookies.jar = cookie_jar
        except (pickle.PickleError, binascii.Error) as e:
            raise exceptions.CookieDecodingError(info=f'Session ID: {self.session_id} -> Cookie Error: {str(e)}')


class SessionPersistClient(RetryClient):
    def __enter__(self):
        services.rehydrate_session_from_database(self)
        return super().__enter__()

    def __exit__(self, *args, **kwargs):
        services.write_session_to_database(self)
        super().__exit__(*args, **kwargs)



class CoreException(Exception):
    ERROR_MESSAGE = 'A generic error has occurred'

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def __str__(self):
        return self.ERROR_MESSAGE

# Boto3 Exceptions Located below:
# https://boto3.amazonaws.com/v1/documentation/api/latest/guide/error-handling.html#botocore-exceptions
# https://github.com/boto/botocore/blob/develop/botocore/exceptions.py


class APIException(CoreException):
    ERROR_MESSAGE = 'A generic API error occurred'


class NotAuthorizedException(APIException):
    ERROR_MESSAGE = 'Client is not authorized to take action'


class CookieDecodingError(APIException):
    ERROR_MESSAGE = 'Error while decoding binary cookies'


class MissingCookieException(APIException):
    ERROR_MESSAGE = 'Missing Cookie Exception'


class AWSCoreException(CoreException):
    ERROR_MESSAGE = 'A generic AWS error occurred'


class CognitoException(AWSCoreException):
    ERROR_MESSAGE = 'An error occurred while attempting to access Cognito'


class DynamoDBException(AWSCoreException):
    ERROR_MESSAGE = 'An error occurred while attempting to access Dynamo DB'


class DBConditionCheckFailed(DynamoDBException):
    ERROR_MESSAGE = 'Condition Check Failed'


class DBTransactionCanceledException(DynamoDBException):
    ERROR_MESSAGE = 'Transaction cancelled'


class SecretsManagerException(AWSCoreException):
    ERROR_MESSAGE = 'An error occurred while fetching secrets'


ERR_CODE_MAP = {
    'ConditionalCheckFailedException': DBConditionCheckFailed,
    'TransactionCanceledException': DBTransactionCanceledException
}



import typing
from abc import ABC

import boto3
from botocore.config import Config

from py_aws_core import logs
from py_aws_core.secrets_manager import get_secrets_manager

logger = logs.logger
secrets_manager = get_secrets_manager()

COGNITO_CLIENT_CONNECT_TIMEOUT = 4.9
COGNITO_CLIENT_READ_TIMEOUT = 4.9


class CognitoClient:
    __CONFIG = Config(
        connect_timeout=COGNITO_CLIENT_CONNECT_TIMEOUT,
        read_timeout=COGNITO_CLIENT_READ_TIMEOUT,
        retries=dict(
            total_max_attempts=2,
        )
    )
    __boto3_session = boto3.Session()

    def __init__(self):
        self._boto_client = None

    @property
    def boto_client(self):
        if not self._boto_client:
            self._boto_client = self.get_new_client()
        return self._boto_client

    @boto_client.setter
    def boto_client(self, value):
        self._boto_client = value

    @property
    def aws_cognito_pool_client_id(self):
        return secrets_manager.get_secret(secret_name='AWS_COGNITO_POOL_CLIENT_ID')

    @property
    def aws_cognito_pool_id(self):
        return secrets_manager.get_secret(secret_name='AWS_COGNITO_POOL_ID')

    @classmethod
    def get_new_client(cls):
        logger.info(f'Getting new Cognito client')
        return cls.__boto3_session.client(
            config=cls.__CONFIG,
            service_name='cognito-idp',
        )

    def admin_create_user(self, *args, **kwargs):
        return self.boto_client.admin_create_user(*args, **kwargs)

    def initiate_auth(self, *args, **kwargs):
        return self.boto_client.initiate_auth(*args, **kwargs)


class AdminCreateUser:
    class Response:
        class User:
            class MFAOptions:
                def __init__(self, data: dict):
                    self.DeliveryMedium = data.get('DeliveryMedium')
                    self.AttributeName = data.get('AttributeName')

            class Attribute:
                def __init__(self, data: dict):
                    self.Name = data.get('Name')
                    self.Value = data.get('Value')

            def __init__(self, data: dict):
                self.Username = data.get('Username')
                self.Attributes = [self.Attribute(a) for a in data.get('Attributes')]
                self.UserCreateDate = data.get('UserCreateDate')
                self.UserLastModifiedDate = data.get('UserLastModifiedDate')
                self.Enabled = data.get('Enabled')
                self.UserStatus = data.get('UserStatus')
                self.MFAOptions = [self.MFAOptions(mfa) for mfa in data.get('MFAOptions')]

        def __init__(self, data: dict):
            self.User = self.User(data.get('User', dict()))

    @classmethod
    def call(
        cls,
        client: CognitoClient,
        username: str,
        user_attributes: typing.List[typing.Dict],
        desired_delivery_mediums: typing.List[str],
    ):
        response = client.admin_create_user(
            DesiredDeliveryMediums=desired_delivery_mediums,
            Username=username,
            UserAttributes=user_attributes,
            UserPoolId=client.aws_cognito_pool_id()
        )
        return cls.Response(response)


class ABCInitiateAuth(ABC):
    class Response:
        class AuthenticationResult:
            class NewDeviceMetadata:
                def __init__(self, data: dict):
                    self.DeviceKey = data.get('DeviceKey', dict())
                    self.DeviceGroupKey = data.get('DeviceGroupKey', dict())

            def __init__(self, data: dict):
                self._data = data
                self.AccessToken = data.get('AccessToken')
                self.ExpiresIn = data.get('ExpiresIn')
                self.TokenType = data.get('TokenType')
                self.RefreshToken = data.get('RefreshToken')
                self.IdToken = data.get('IdToken')
                self.NewDeviceMetadata = self.NewDeviceMetadata(data.get('NewDeviceMetadata', dict()))

        def __init__(self, data: dict):
            self.ChallengeName = data.get('ChallengeName')
            self.Session = data.get('Session')
            self.ChallengeParameters = data.get('ChallengeParameters')
            self.AuthenticationResult = self.AuthenticationResult(data['AuthenticationResult'])


class UserPasswordAuth(ABCInitiateAuth):
    @classmethod
    def call(
        cls,
        client: CognitoClient,
        username: str,
        password: str,
    ):
        response = client.initiate_auth(
            AuthFlow='USER_PASSWORD_AUTH',
            AuthParameters={
                'USERNAME': username,
                'PASSWORD': password,
            },
            ClientId=client.aws_cognito_pool_client_id(),
        )
        return cls.Response(response)


class RefreshTokenAuth(ABCInitiateAuth):
    @classmethod
    def call(
        cls,
        client: CognitoClient,
        refresh_token: str,
    ):
        response = client.initiate_auth(
            AuthFlow='REFRESH_TOKEN',
            AuthParameters={
                'REFRESH_TOKEN': refresh_token,
            },
            ClientId=client.aws_cognito_pool_client_id(),
        )
        return cls.Response(response)



import logging

from httpx import Client

from py_aws_core import decorators
from . import exceptions

logger = logging.getLogger(__name__)


STATUS_ERRORS_MAP = {
    407: exceptions.ProxyNotAuthenticated,
    560: exceptions.GeoLocationNotFound,
    561: exceptions.ProxyUnreachable,
    562: exceptions.ProxyNotFound,
    564: exceptions.ProxyNotOnline
}


class ProxyrackAPI:
    ROOT_URL = 'http://api.proxyrack.net'


class GetActiveConnections(ProxyrackAPI):
    class Response:
        class Connection:
            def __init__(self, data):
                self.create_time = data['createTime']
                self.dest_addr = data['destAddr']
                self.source_addr = data['sourceAddr']

        def __init__(self, data):
            self.connections = [self.Connection(c) for c in data]

    @classmethod
    @decorators.wrap_exceptions(raise_as=exceptions.ProxyRackException)
    def call(cls, client: Client, *args, **kwargs):
        url = f'{cls.ROOT_URL}/active_conns'
        r = client.get(url=url, *args, **kwargs)
        r.raise_for_status()
        return cls.Response(r.json())


class GetCities(ProxyrackAPI):
    class Response:
        def __init__(self, data):
            self.cities = data

    @classmethod
    @decorators.wrap_exceptions(raise_as=exceptions.ProxyRackException)
    def call(cls, client: Client, *args, **kwargs):
        url = f'{cls.ROOT_URL}/cities'
        r = client.get(url=url, *args, **kwargs)
        r.raise_for_status()
        return cls.Response(r.json())


class GetCountries(ProxyrackAPI):
    class Response:
        def __init__(self, data):
            self.countries = data

    @classmethod
    @decorators.wrap_exceptions(raise_as=exceptions.ProxyRackException)
    def call(cls, client: Client, *args, **kwargs):
        url = f'{cls.ROOT_URL}/countries'
        r = client.get(url=url, *args, **kwargs)
        r.raise_for_status()
        return cls.Response(r.json())


class GetISPs(ProxyrackAPI):
    class Response:
        def __init__(self, data):
            self.isps = data

    @classmethod
    @decorators.wrap_exceptions(raise_as=exceptions.ProxyRackException)
    def call(cls, client: Client, country: str, *args, **kwargs):
        url = f'{cls.ROOT_URL}/countries/{country}/isps'
        r = client.get(url=url, *args, **kwargs)
        r.raise_for_status()
        return cls.Response(r.json())


class GetCountryIPCount(ProxyrackAPI):
    @classmethod
    @decorators.wrap_exceptions(raise_as=exceptions.ProxyRackException)
    def call(cls, client: Client, country: str, *args, **kwargs):
        url = f'{cls.ROOT_URL}/countries/{country}/count'
        r = client.get(url=url, *args, **kwargs)
        r.raise_for_status()
        return r.text


class PostTempAPIKey(ProxyrackAPI):
    class Response:
        class Password:
            def __init__(self, data):
                self.expiration_seconds = data['expirationSeconds']
                self.password = data['password']

        def __init__(self, data):
            self.password = self.Password(data['password'])
            self.success = data['success']

        @property
        def api_key(self):
            return self.password.password

    @classmethod
    @decorators.wrap_exceptions(raise_as=exceptions.ProxyRackException)
    def call(cls, client: Client, expiration_seconds: int, *args, **kwargs):
        url = f'{cls.ROOT_URL}/passwords'

        params = {
            'expirationSeconds': expiration_seconds,
        }

        r = client.post(url=url, params=params, *args, **kwargs)
        r.raise_for_status()
        return cls.Response(r.json())


class GetStats(ProxyrackAPI):
    class Response:
        class IPInfo:
            class Fingerprint:
                def __init__(self, data):
                    self.osName = data['osName']

            def __init__(self, data):
                self.city = data['city']
                self.country = data['country']
                self.fingerprint = self.Fingerprint(data['fingerprint'])
                self.ip = data['ip']
                self.isp = data['isp']
                self.online = data['online']
                self.proxyId = data.get('proxyId')

        def __init__(self, data):
            self.ipinfo = self.IPInfo(data['ipinfo'])
            self.thread_limit = data['threadLimit']

    @classmethod
    @decorators.wrap_exceptions(raise_as=exceptions.ProxyRackException)
    def call(cls, client: Client, *args, **kwargs):
        """
            Gets statistics of current proxy
            Note: An endpoint must be hit first, or data will be blank
            :param client:
            :param args:
            :param kwargs:
            :return:
        """

        url = f'{cls.ROOT_URL}/stats'
        r = client.get(url=url, *args, **kwargs)
        r.raise_for_status()
        return cls.Response(r.json())



import typing
from abc import ABC

import boto3
from botocore.config import Config

from py_aws_core import const, entities, logs, utils
from .secrets_manager import get_secrets_manager

secrets_manager = get_secrets_manager()

logger = logs.logger

DDB_CLIENT_CONNECT_TIMEOUT = 4.9
DDB_CLIENT_READ_TIMEOUT = 4.9


class DDBClient:
    __config = Config(
        connect_timeout=DDB_CLIENT_CONNECT_TIMEOUT,
        read_timeout=DDB_CLIENT_READ_TIMEOUT,
        tcp_keepalive=True,  # Keep TCP connections open for faster response
        retries=dict(
            total_max_attempts=2,
        )
    )
    __boto3_session = boto3.Session()

    def __init__(self):
        self._boto_client = None
        self._table_resource = None

    @property
    def boto_client(self):
        if not self._boto_client:
            self._boto_client = self.get_new_client()
        return self._boto_client

    @boto_client.setter
    def boto_client(self, value):
        self._boto_client = value

    @property
    def table_resource(self):
        if not self._table_resource:
            self._table_resource = self.get_new_table_resource()
        return self._table_resource

    @classmethod
    def get_new_client(cls):
        logger.info(f'Getting new DynamoDB client')
        return cls.__boto3_session.client(
            config=cls.__config,
            service_name='dynamodb',
            verify=False  # Don't validate SSL certs for faster responses
        )

    @classmethod
    def get_table_name(cls):
        return secrets_manager.get_secret(secret_name='AWS_DYNAMO_DB_TABLE_NAME')

    def get_new_table_resource(self):
        dynamodb = boto3.resource('dynamodb')
        return dynamodb.Table(self.get_table_name())

    def query(
        self,
        *args,
        key_condition_expression: str,
        expression_attribute_names: typing.Dict,
        expression_attribute_values: typing.Dict,
        projection_expression: str = None,
        **kwargs
    ):
        return self.boto_client.query(
            TableName=self.get_table_name(),
            KeyConditionExpression=key_condition_expression,
            ExpressionAttributeNames=expression_attribute_names,
            ExpressionAttributeValues=expression_attribute_values,
            ProjectionExpression=projection_expression,
            *args,
            **kwargs
        )

    def scan(self, *args, **kwargs):
        return self.boto_client.scan(TableName=self.get_table_name(), *args, **kwargs)

    def get_item(self, *args, **kwargs):
        return self.boto_client.get_item(TableName=self.get_table_name(), *args, **kwargs)

    def put_item(self, *args, **kwargs):
        return self.boto_client.put_item(TableName=self.get_table_name(), *args, **kwargs)

    def delete_item(self, *args, **kwargs):
        return self.boto_client.delete_item(TableName=self.get_table_name(), *args, **kwargs)

    def update_item(
        self,
        *args,
        key: typing.Dict,
        update_expression: str,
        expression_attribute_values: typing.Dict,
        **kwargs
    ):
        return self.boto_client.update_item(
            *args,
            TableName=self.get_table_name(),
            Key=key,
            UpdateExpression=update_expression,
            ExpressionAttributeValues=expression_attribute_values,
            **kwargs
        )

    def batch_write_item(self, *args, **kwargs):
        return self.boto_client.batch_write_item(*args, **kwargs)

    def transact_write_items(self, *args, **kwargs):
        return self.boto_client.transact_write_items(*args, **kwargs)

    def batch_write_item_maps(self, item_maps: typing.List[typing.Dict]) -> int:
        with self.table_resource.batch_writer() as batch:
            for _map in item_maps:
                batch.put_item(Item=_map)
        return len(item_maps)

    def write_maps_to_db(self, item_maps: typing.List[typing.Dict]) -> int:
        return self.batch_write_item_maps(item_maps=item_maps)


__db_client = DDBClient()


def get_db_client():
    """
    Reuses db client across all modules for efficiency
    :return:
    """
    return __db_client


class QueryResponse(ABC):
    def __init__(self, data):
        self.Items = data.get('Items')
        self.Count = data.get('Count')
        self.ScannedCount = data.get('ScannedCount')
        self.ResponseMetadata = ResponseMetadata(data['ResponseMetadata'])

    def get_by_type(self, _type: str) -> typing.List:
        if self.Items:
            return [item for item in self.Items if entities.BaseModel(item).Type == _type]
        return list()


class ABCCommonAPI(ABC):
    """
        https://docs.aws.amazon.com/amazondynamodb/latest/developerguide/Expressions.ConditionExpressions.html
        If your primary key consists of both a partition key(pk) and a sort key(sk),
        the parameter will check whether attribute_not_exists(pk) AND attribute_not_exists(sk) evaluate to true or
        false before attempting the write operation.
    """
    @classmethod
    def get_batch_entity_create_map(
        cls,
        pk: str,
        sk: str,
        _type: str,
        created_by: str = '',
        expire_in_seconds: int | None = const.DB_DEFAULT_EXPIRES_IN_SECONDS,
        **kwargs,
    ):
        return {
            'PK': pk,
            'SK': sk,
            'Type': _type,
            'CreatedAt': cls.iso_8601_now_timestamp(),
            'CreatedBy': created_by,
            'ModifiedAt': '',
            'ModifiedBy': '',
            'ExpiresAt': cls.calc_expire_at_timestamp(expire_in_seconds=expire_in_seconds),
        } | kwargs

    @classmethod
    def get_entity_update_map(
        cls,
        pk: str,
        sk: str,
        modified_by: str = '',
        **kwargs,
    ):
        return {
            'PK': pk,
            'SK': sk,
            'ModifiedAt': cls.iso_8601_now_timestamp(),
            'ModifiedBy': modified_by,
        } | kwargs

    @classmethod
    def iso_8601_now_timestamp(cls) -> str:
        return utils.to_iso_8601()

    @classmethod
    def calc_expire_at_timestamp(cls, expire_in_seconds: int = None) -> int | str:
        """
        Adds seconds to current unix timestamp to generate new unix timestamp
        Seconds set to None will result in empty string
        :param expire_in_seconds: Seconds to add to current timestamp
        :return:
        """
        if expire_in_seconds is None:
            return ''
        return utils.add_seconds_to_current_unix_timestamp(seconds=expire_in_seconds)


class ErrorResponse:
    class Error:
        def __init__(self, data):
            self.Message = data['Message']
            self.Code = data['Code']

    class CancellationReason:
        def __init__(self, data):
            self.Code = data['Code']
            self.Message = data.get('Message')

    def __init__(self, data):
        self.Error = self.Error(data.get('Error', dict()))
        self.ResponseMetadata = ResponseMetadata(data.get('ResponseMetadata', dict()))
        self.Message = data.get('Message')
        self.CancellationReasons = [self.CancellationReason(r) for r in data.get('CancellationReasons', list())]

    def raise_for_cancellation_reasons(self, error_maps: typing.List[typing.Dict[str, typing.Any]]):
        for reason, error_map in zip(self.CancellationReasons, error_maps):
            if exc := error_map.get(reason.Code):
                raise exc


class DDBItemResponse(ABC):
    def __init__(self, data):
        self._data = data
        self.Type = data.get('__type')
        self.Item = data.get('Item')
        self.ResponseMetadata = ResponseMetadata(data.get('ResponseMetadata', dict()))


class DynamoDBTransactResponse(ABC):
    def __init__(self, data):
        self._data = data
        self.Responses = data.get('Responses')

    @property
    def data(self):
        return self._data


class ResponseMetadata:
    class HTTPHeaders:
        def __init__(self, data):
            self.server = data.get('server')
            self.date = data.get('date')

    def __init__(self, data):
        self.RequestId = data.get('RequestId')
        self.HTTPStatusCode = data.get('HTTPStatusCode')
        self.HTTPHeaders = self.HTTPHeaders(data.get('HTTPHeaders', dict()))
        self.RetryAttempts = data.get('RetryAttempts')



import json

import boto3
from botocore.exceptions import ClientError

from py_aws_core import exceptions, logs, utils

logger = logs.logger


class SecretsManager:
    """
    First checks environment variables for secrets.
    If secret not found, will attempt to pull from secrets manager
    """
    AWS_SECRET_NAME = 'AWS_SECRET_NAME'

    def __init__(self):
        self._boto_client = None
        self._secrets_map = dict()

    def get_secret(self, secret_name: str):
        if secret_value := utils.get_environment_variable(secret_name):
            logger.debug(f'Secret "{secret_name}" found in environment variables')
            return secret_value
        if val := self._secrets_map.get(secret_name):
            logger.debug(f'Secret "{secret_name}" found in cached secrets')
            return val
        try:
            r_secrets = self.boto_client.get_secret_value(SecretId=self.get_aws_secret_id)
            self._secrets_map = json.loads(r_secrets['SecretString'])
            return self._secrets_map[secret_name]
        except ClientError as e:
            logger.exception(f'Error while trying to find secret "{secret_name}"')
            # For a list of exceptions thrown, see
            # https://docs.aws.amazon.com/secretsmanager/latest/apireference/API_GetSecretValue.html
            raise exceptions.SecretsManagerException(e)

    @property
    def boto_client(self):
        if not self._boto_client:
            self._boto_client = boto3.client('secretsmanager')
        return self._boto_client

    @boto_client.setter
    def boto_client(self, value):
        self._boto_client = value

    @property
    def get_aws_secret_id(self) -> str:
        if aws_secret_id := utils.get_environment_variable(self.AWS_SECRET_NAME):
            return aws_secret_id
        raise exceptions.SecretsManagerException(f'Missing environment variable "{self.AWS_SECRET_NAME}"')


__secrets_manager = SecretsManager()


def get_secrets_manager() -> SecretsManager:
    """
    Reuses secrets manager across all modules for efficiency
    :return:
    """
    return __secrets_manager



import logging
import random
from abc import ABC, abstractmethod

from httpx import Client

from py_aws_core.secrets_manager import get_secrets_manager
from . import const

logger = logging.getLogger(__name__)
secrets_manager = get_secrets_manager()


class ProxyBackend(ABC):
    @classmethod
    @abstractmethod
    def get_proxy_url(cls, **kwargs) -> str:
        raise NotImplemented

    @staticmethod
    def get_weighted_country():
        countries, weights = zip(const.PROXY_COUNTRY_WEIGHTS)
        return random.choices(population=countries, weights=weights, k=1)[0]

    @classmethod
    def get_proxy_password(cls):
        return secrets_manager.get_secret(secret_name='PROXY_PASSWORD')

    @classmethod
    def get_proxy_username(cls):
        return secrets_manager.get_secret(secret_name='PROXY_USERNAME')



import json
import os
import random
import time
import typing
import uuid
from datetime import datetime, timezone, timedelta, UTC


def build_lambda_response(
    status_code: int,
    body: typing.Dict | str = None,
    multi_value_headers: typing.Dict[str, typing.List[str]] = None,
    exc: Exception = None
):
    """
      Note - CORS response headers will NOT be present in response if using proxy type integration.
      CORS response headers must be set in lambda response
      See below for further details:
      https://stackoverflow.com/a/54089431
      https://docs.aws.amazon.com/apigateway/latest/developerguide/how-to-cors.html#apigateway-enable-cors-proxy
    :param multi_value_headers:
    :param status_code:
    :param body:
    :param exc:
    :return:
    """
    if not body:
        body = dict()
    if exc:
        body['error'] = f'{type(exc).__name__}: {str(exc)}'
    if isinstance(body, dict):
        body = json.dumps(body)
    response_headers = {
        'Content-Type': ['application/json'],
        'Access-Control-Allow-Credentials': [True],
        'Access-Control-Allow-Headers': ['Content-Type,X-Amz-Date,Authorization,X-Api-Key,X-Amz-Security-Token'],
        'Access-Control-Allow-Origin': ['*'],
        'Access-Control-Allow-Methods': ['DELETE,GET,POST,PUT']
    }
    if multi_value_headers:
        response_headers |= multi_value_headers
    return {
        'isBase64Encoded': False,
        'statusCode': status_code,
        'body': body,  # body key value must be a json string
        'multiValueHeaders': response_headers
    }


def to_iso_8601(dt: datetime = None, tz=timezone.utc) -> str:
    """
    Example output 2020-07-10 15:00:00.000
    :param tz:
    :param dt:
    :return:
    """
    if not dt:
        dt = get_now_datetime(tz=tz).replace(microsecond=0)
    return dt.isoformat()


def add_seconds_to_current_unix_timestamp(seconds: int, tz=timezone.utc) -> int:
    dt = get_now_datetime(tz=tz) + timedelta(seconds=seconds)
    return int(dt.timestamp())


def add_seconds_to_now_datetime(seconds: int, tz=timezone.utc) -> datetime:
    return get_now_datetime(tz=tz) + timedelta(seconds=seconds)


def get_now_datetime(tz=timezone.utc) -> datetime:
    return datetime.now(tz=tz)


def generate_jitter(midpoint: float, floor: float, std_deviation: float) -> float:
    return max(floor, midpoint + random.uniform(-std_deviation, +std_deviation))


def rand_int(num_a: int, num_b: int) -> int:
    return random.randint(num_a, num_b)


def sleep(seconds: float) -> None:
    return time.sleep(seconds)


def get_uuid() -> uuid.UUID:
    return uuid.uuid4()


def get_environment_variable(secret_name: str, default=None) -> str:
    return os.environ.get(secret_name, default=default)


def decode_unicode(s: str) -> str:
    """
    Per https://stackoverflow.com/a/33134946
    This is how to fix unicode decoding errors
    :param s: string to decode
    :return:
    """
    try:
        return s.encode('iso-8859-1').decode('utf-8')
    except (UnicodeEncodeError, UnicodeDecodeError):
        return s


def unix_timestamp_to_iso8601(unix_ts: int):
    return datetime.fromtimestamp(unix_ts, tz=UTC).isoformat()


def remove_whitespace(s: str) -> str:
    return ''.join([line.replace(' ', '') for line in s.split('\\r\\n')])



import uuid
from abc import ABC, abstractmethod

from boto3.dynamodb import types

from py_aws_core import encoders


class BaseModel:
    def __init__(self, data):
        self.__data = self.to_normalized_data(data)
        self.PK = self.data.get('PK')
        self.SK = self.data.get('SK')
        self.Type = self.data.get('Type')
        self.CreatedAt = self.data.get('CreatedAt')
        self.CreatedBy = self.data.get('CreatedBy')
        self.ModifiedAt = self.data.get('ModifiedAt')
        self.ModifiedBy = self.data.get('ModifiedBy')
        self.ExpiresAt = self.data.get('ExpiresAt')

    @property
    def data(self):
        return self.__data

    @staticmethod
    def to_normalized_data(data: dict) -> dict:
        """
        Converts low level dynamo json to normalized json
        """
        return {k: types.TypeDeserializer().deserialize(v) for k, v in data.items()}

    @property
    def to_json(self):
        return encoders.DBEncoder().serialize_to_json(self)


class ABCEntity(ABC, BaseModel):
    TYPE = 'ABC'

    @classmethod
    @abstractmethod
    def create_key(cls, *args, **kwargs) -> str:
        pass

    @classmethod
    def type(cls) -> str:
        return cls.TYPE


class Session(ABCEntity):
    TYPE = 'SESSION'

    def __init__(self, data):
        super().__init__(data)
        self.Base64Cookies = self.data['Base64Cookies']

    @classmethod
    def create_key(cls, _id: uuid.UUID) -> str:
        return f'{cls.type()}#{str(_id)}'



import time
from http.cookiejar import Cookie
from unittest import TestCase

import respx
from httpx import Response, codes


class BaseTestFixture(TestCase):
    def setUp(self):
        self.start_time = time.time()
        super().setUp()

    def tearDown(self):
        total = time.time() - self.start_time
        if total > 0.1:
            print(f'{self.id} slow test took {total:.3f} seconds')
        super().tearDown()

    @classmethod
    def create_ok_route(cls, method='GET', headers=None, _json=None, text=None, **kwargs):
        return cls.create_route(
            method=method,
            response_status_code=codes.OK,
            response_headers=headers,
            response_json=_json,
            response_text=text,
            **kwargs
        )

    @classmethod
    def create_bad_request_route(cls, **kwargs):
        return cls.create_route(response_status_code=codes.BAD_REQUEST, **kwargs)

    @classmethod
    def create_route(
        cls,
        response_status_code,
        method='GET',
        response_headers=None,
        response_json=None,
        response_text=None,
        **kwargs
    ):
        return respx.route(
            method=method,
            **kwargs
        ).mock(
            return_value=Response(
                headers=response_headers,
                status_code=response_status_code,
                json=response_json,
                text=response_text,
            )
        )

    @classmethod
    def create_test_cookie(cls, name: str, value: str) -> Cookie:
        return Cookie(
            version=1,
            name=name,
            value=value,
            port=None,
            port_specified=False,
            domain='www.example.com',
            domain_specified=True,
            domain_initial_dot=False,
            path='/',
            path_specified=True,
            secure=True,
            expires=3600,
            discard=False,
            comment=None,
            comment_url=None,
            rest=dict(),
        )



import logging
import typing
import uuid

from py_aws_core.spoofing.backends import ProxyBackend
from py_aws_core.spoofing.proxyrack import const, utils

logger = logging.getLogger(__name__)


class ProxyRackProxyBackend(ProxyBackend):
    @classmethod
    def get_proxy_url(
        cls,
        netloc: str,
        cities: typing.List[str] = None,
        country: str = None,
        isps: typing.List[str] = None,
        proxy_ip: str = None,
        proxy_os: const.ProxyOs = None,
        session_id: str | uuid.UUID = None,
        **kwargs
    ) -> str:
        config = utils.ProxyBuilder.Config(
            cities=cities,
            country=country,
            isps=isps,
            proxy_ip=proxy_ip,
            proxy_os=proxy_os,
            session_id=session_id,
            refresh_minutes=60
        )
        return utils.ProxyBuilder(
            username=cls.get_proxy_username(),
            password=cls.get_proxy_password(),
            netloc=netloc,
            config=config
        ).http_url



import logging.config
import os

from dotenv import load_dotenv

dotenv = load_dotenv()  # take environment variables from .env.

logger = logging.getLogger(__name__)


class Logging:
    @classmethod
    def get_logging_config(cls) -> dict:
        return {
            'version': 1,
            'disable_existing_loggers': False,
            'formatters': {
                'simple': {
                    'format': '[{levelname}] {message}',
                    'style': '{',
                },
                'verbose': {
                    'format': '[{asctime} {levelname}/{threadName}] {message}',
                    'style': '{',
                },
                'verbose_plus': {
                    'format': '[{asctime} {levelname}/{threadName} {pathname}] {message}',
                    'style': '{',
                },
            },
            'handlers': {
                'console': {
                    'level': 'DEBUG',
                    'class': 'logging.StreamHandler',
                    'formatter': 'verbose',
                },
            },
            'loggers': {
                '': {
                    'handlers': ['console'],
                    'level': cls.get_log_level(),
                    'propagate': True,
                }
            },
        }

    @classmethod
    def get_log_level(cls):
        log_level_upper = os.environ.get('LOG_LEVEL', 'DEBUG').upper()
        return getattr(logging, log_level_upper)

    @classmethod
    def configure_logger(cls):
        logging.config.dictConfig(Logging.get_logging_config())


Logging.configure_logger()
